from ocpa.visualization.oc_petri_net.versions import control_flow, annotated_with_diagnostics
from pm4py.visualization.common import gview
from pm4py.visualization.common import save as gsave


CONTROL_FLOW = "control_flow"
ANNOTATED_WITH_DIAGNOSTICS = "annotated_with_diagnostics"

VERSIONS = {
    CONTROL_FLOW: control_flow.apply,
    ANNOTATED_WITH_DIAGNOSTICS: annotated_with_diagnostics.apply
}


def apply(obj, variant=CONTROL_FLOW, **kwargs):
    return VERSIONS[variant](obj, **kwargs)


def save(gviz, output_file_path):
    """
    Save the diagram

    Parameters
    -----------
    gviz
        GraphViz diagram
    output_file_path
        Path where the GraphViz output should be saved
    """
    gsave.save(gviz, output_file_path)


def view(gviz):
    """
    View the diagram

    Parameters
    -----------
    gviz
        GraphViz diagram
    """
    return gview.view(gviz)



import uuid
import tempfile
from graphviz import Digraph
from ocpa.objects.oc_petri_net.obj import ObjectCentricPetriNet

COLORS = ["#05B202", "#A13CCD", "#BA0D39", "#39F6C0", "#E90638", "#07B423", "#306A8A", "#678225", "#2742FE", "#4C9A75",
          "#4C36E9", "#7DB022", "#EDAC54", "#EAC439", "#EAC439", "#1A9C45", "#8A51C4", "#496A63", "#FB9543", "#2B49DD",
          "#13ADA5", "#2DD8C1", "#2E53D7", "#EF9B77", "#06924F", "#AC2C4D", "#82193F", "#0140D3"]


def apply(obj, parameters=None):
    if parameters is None:
        parameters = {}

    image_format = "png"
    if "format" in parameters:
        image_format = parameters["format"]

    filename = tempfile.NamedTemporaryFile(suffix='.gv').name
    g = Digraph("", filename=filename, engine='dot',
                graph_attr={'bgcolor': 'transparent'})

    all_objs = {}
    trans_names = {}

    pl_count = 1
    tr_count = 1
    arc_count = 1

    # color = COLORS[index % len(COLORS)]
    color = "#05B202"
    color_mapping = dict()
    object_types = obj.object_types
    for index, ot in enumerate(object_types):
        color_mapping[ot] = COLORS[index % len(COLORS)]

    for pl in obj.places:
        # this_uuid = str(uuid.uuid4())
        this_uuid = "p%d" % (pl_count)
        # pl_str = this_uuid
        pl_count += 1
        color = color_mapping[pl.object_type]
        if pl.initial == True:
            g.node(pl.name, pl.name, shape="circle", style="filled", fillcolor=color, color=color,
                   fontsize="13.0", labelfontsize="13.0")
        elif pl.final == True:
            g.node(pl.name, pl.name, shape="circle", style="filled", color=color, fillcolor=color,
                   fontsize="13.0", labelfontsize="13.0")
        else:
            g.node(pl.name, pl.name, shape="circle", color=color,
                   fontsize="13.0", labelfontsize="13.0")
        all_objs[pl] = pl.name

    for tr in obj.transitions:
        # this_uuid = str(uuid.uuid4())
        this_uuid = "t%d" % (tr_count)
        tr_count += 1
        if tr.silent == True:
            g.node(this_uuid, this_uuid, fontcolor="#FFFFFF", shape="box",
                   fillcolor="#000000", style="filled", xlabel="Test", labelfontsize="18.0")
            all_objs[tr] = this_uuid
        elif tr.name not in trans_names:
            g.node(this_uuid, "%s \n (%s)" % (tr.name, this_uuid), shape="box", fontsize="36.0",
                   labelfontsize="36.0")
            trans_names[tr.name] = this_uuid
            all_objs[tr] = this_uuid
        else:
            all_objs[tr] = trans_names[tr.name]

    for arc in obj.arcs:
        # this_uuid = str(uuid.uuid4())
        this_uuid = "a%d" % (arc_count)
        arc_count += 1

        source_node = arc.source
        target_node = arc.target

        if type(source_node) == ObjectCentricPetriNet.Place:
            color = color_mapping[source_node.object_type]
        if type(target_node) == ObjectCentricPetriNet.Place:
            color = color_mapping[target_node.object_type]

        if arc.variable == True:
            g.edge(all_objs[source_node], all_objs[target_node],
                   label="", color=color + ":white:" + color, fontsize="13.0")
        else:
            g.edge(all_objs[source_node], all_objs[target_node],
                   label="", color=color, fontsize="13.0")

        all_objs[arc] = this_uuid

    g.attr(overlap='false')
    g.attr(fontsize='11')

    g.format = image_format
    return g



import pm4py

from pm4py.algo.discovery.alpha import algorithm as alpha_miner
from pm4py.algo.discovery.inductive import algorithm as inductive_miner
from pm4py.algo.discovery.dfg import algorithm as dfg_discovery
from pm4py.statistics.start_activities.log import get as sa_get
from pm4py.statistics.end_activities.log import get as ea_get
from pm4py.objects.conversion.dfg import converter as dfg_converter
from pm4py.algo.conformance.tokenreplay import algorithm as tr_factory
from pm4py.visualization.petrinet.util import performance_map
from pm4py.statistics.variants.log import get as variants_module
from pm4py.algo.filtering.log.paths import paths_filter
from pm4py.algo.filtering.log.variants import variants_filter
from pm4py.algo.filtering.log.attributes import attributes_filter
from pm4py.objects.petri.petrinet import PetriNet, Marking
from pm4py.objects.petri.utils import add_arc_from_to
from pm4py.objects.petri.utils import remove_place, remove_transition
from ocpa.objects.log.importer.mdl import factory as mdl_import_factory
from ocpa.algo.discovery.mvp.projection import algorithm
from ocpa.objects.oc_petri_net.obj import ObjectCentricPetriNet
from copy import deepcopy
import uuid
import pandas as pd
import time

PARAM_ACTIVITY_KEY = pm4py.util.constants.PARAMETER_CONSTANT_ACTIVITY_KEY
MAX_FREQUENCY = float("inf")


def reduce_petri_net(net):
    transes = set([x for x in net.transitions if x.label is None])
    places = list(net.places)
    i = 0
    while i < len(places):
        place = places[i]
        source_transes = set([x.source for x in place.in_arcs])
        target_transes = set([x.target for x in place.out_arcs])
        if len(source_transes) == 1 and len(target_transes) == 1:
            if source_transes.issubset(transes) and target_transes.issubset(transes):
                source_trans = list(source_transes)[0]
                target_trans = list(target_transes)[0]
                if len(target_trans.out_arcs) == 1:
                    target_place = list(target_trans.out_arcs)[0].target
                    add_arc_from_to(source_trans, target_place, net)
                    remove_place(net, place)
                    remove_transition(net, target_trans)
                    places = list(net.places)
                    continue
                    # print(source_trans, target_trans, target_place)
        i = i + 1

    return net


def discover_alpha(log):
    return alpha_miner.apply(log)


def discover_inductive(log):
    return inductive_miner.apply(log)


def discover_dfg_miner(log):
    dfg = dfg_discovery.apply(log)
    sa = sa_get.get_start_activities(log)
    ea = ea_get.get_end_activities(log)
    net, im, fm = dfg_converter.apply(
        dfg, parameters={"start_activities": sa, "end_activities": ea})
    return net, im, fm


def discover_nets(df, discovery_algorithm=discover_inductive, parameters=None):
    if parameters is None:
        parameters = {}

    allowed_activities = parameters["allowed_activities"] if "allowed_activities" in parameters else None
    debug = parameters["debug"] if "debug" in parameters else True

    # try:
    #     if df.type == "succint":
    #         df = mdl_import_factory.succint_mdl_to_exploded_mdl(df)
    #         df.type = "exploded"
    # except:
    #     pass
    df = mdl_import_factory.succint_mdl_to_exploded_mdl(df)

    if len(df) == 0:
        df = pd.DataFrame({"event_id": [], "event_activity": []})

    min_node_freq = parameters["min_node_freq"] if "min_node_freq" in parameters else 0
    min_edge_freq = parameters["min_edge_freq"] if "min_edge_freq" in parameters else 0

    df = mdl_import_factory.clean_frequency(df, min_node_freq)
    df = mdl_import_factory.clean_arc_frequency(df, min_edge_freq)

    if len(df) == 0:
        df = pd.DataFrame({"event_id": [], "event_activity": []})

    persps = [x for x in df.columns if not x.startswith("event_")]

    ret = {}
    ret["nets"] = {}
    ret["group_size_hist"] = {}

    diff_log = 0
    diff_model = 0
    diff_token_replay = 0
    diff_performance_annotation = 0
    diff_basic_stats = 0

    for persp in persps:
        aa = time.time()
        if debug:
            print(persp, "getting log")
        log = algorithm.apply(df, persp, parameters=parameters)
        print(log)
        if debug:
            print(len(log))

        if allowed_activities is not None:
            if persp not in allowed_activities:
                continue
            filtered_log = attributes_filter.apply_events(
                log, allowed_activities[persp])
        else:
            filtered_log = log
        bb = time.time()

        diff_log += (bb - aa)

        # filtered_log = variants_filter.apply_auto_filter(deepcopy(filtered_log), parameters={"decreasingFactor": 0.5})

        if debug:
            print(len(log))
            print(persp, "got log")

        cc = time.time()
        #net, im, fm = inductive_miner.apply(filtered_log)
        net, im, fm = discovery_algorithm(filtered_log)
        dd = time.time()
        diff_log += (dd - cc)

        if debug:
            print(len(log))
            print(persp, "discovered net")

        group_size_hist = algorithm.apply(
            df, persp, variant="group_size_hist", parameters=parameters)

        ret["nets"][persp] = [net, im, fm]
        ret["group_size_hist"][persp] = group_size_hist

    return ret


def apply(df, discovery_algorithm=discover_inductive, parameters=None):
    ret = discover_nets(df, discovery_algorithm, parameters)
    nets = ret["nets"]
    group_size_hist_persp = ret["group_size_hist"]
    transitions = {}
    transition_list = []
    places = []
    arcs = []
    for index, persp in enumerate(nets):
        net, im, fm = nets[persp]
        pl_count = 1
        tr_count = 1
        arc_count = 1
        group_size_hist = group_size_hist_persp[persp]
        for pl in net.places:
            p = None
            p_name = "%s%d" % (persp, pl_count)
            # also change the name of the net to synchronize with the resulting ocpn
            pl.name = p_name
            pl_count += 1
            # check if initial
            if pl in im:
                p = ObjectCentricPetriNet.Place(name=p_name,
                                                object_type=persp, initial=True)
            elif pl in fm:
                p = ObjectCentricPetriNet.Place(
                    name=p_name, object_type=persp, final=True)
            else:
                p = ObjectCentricPetriNet.Place(name=p_name, object_type=persp)
            places.append(p)

            for arc in pl.in_arcs:
                t = None
                if arc.source.label != "" and arc.source.label != None:
                    if arc.source.label not in transitions.keys():
                        t = ObjectCentricPetriNet.Transition(
                            name=arc.source.label)
                        transitions[arc.source.label] = t
                        transition_list.append(t)
                    else:
                        t = transitions[arc.source.label]
                else:
                    if arc.source.name not in transitions.keys():
                        t = ObjectCentricPetriNet.Transition(
                            name=arc.source.name, silent=True)
                        transitions[arc.source.name] = t
                        transition_list.append(t)
                    else:
                        t = transitions[arc.source.name]

                # add arc
                if arc.source.label in group_size_hist and sum(group_size_hist[arc.source.label]) != len(
                        group_size_hist[arc.source.label]):
                    a = ObjectCentricPetriNet.Arc(t, p, variable=True)
                else:
                    a = ObjectCentricPetriNet.Arc(t, p)
                p.in_arcs.add(a)
                t.out_arcs.add(a)
                arcs.append(a)
            for arc in pl.out_arcs:
                t = None
                if arc.target.label != "" and arc.target.label != None:
                    if arc.target.label not in transitions.keys():
                        t = ObjectCentricPetriNet.Transition(
                            name=arc.target.label)
                        transitions[arc.target.label] = t
                        transition_list.append(t)
                    else:
                        t = transitions[arc.target.label]
                else:
                    if arc.target.name not in transitions.keys():
                        t = ObjectCentricPetriNet.Transition(
                            name=arc.target.name, silent=True)
                        transitions[arc.target.name] = t
                        transition_list.append(t)
                    else:
                        t = transitions[arc.target.name]

                # add arc
                if arc.target.label in group_size_hist and sum(group_size_hist[arc.target.label]) != len(group_size_hist[arc.target.label]):
                    a = ObjectCentricPetriNet.Arc(p, t, variable=True)
                else:
                    a = ObjectCentricPetriNet.Arc(p, t)

                p.out_arcs.add(a)
                t.in_arcs.add(a)
                arcs.append(a)
    ocpn = ObjectCentricPetriNet(
        places=places, transitions=transition_list, arcs=arcs, nets=nets)
    return ocpn



from dataclasses import dataclass, field
from typing import List, Dict, Set, Tuple


class ObjectCentricPetriNet(object):
    class Place(object):
        def __init__(self, name, object_type, out_arcs=None, in_arcs=None, initial=False, final=False):
            self.__name = name
            self.__object_type = object_type
            self.__initial = initial
            self.__final = final
            self.__in_arcs = in_arcs if in_arcs != None else set()
            if out_arcs != None:
                self.__out_arcs = out_arcs
            else:
                self.__out_arcs = set()

        def __set_name(self, name):
            self.__name = name

        def __get_name(self):
            return self.__name

        def __get_object_type(self):
            return self.__object_type

        def __get_initial(self):
            return self.__initial

        def __get_final(self):
            return self.__final

        def __get_out_arcs(self):
            return self.__out_arcs

        def __set_out_arcs(self, out_arcs):
            self.__out_arcs = out_arcs

        def __get_in_arcs(self):
            return self.__in_arcs

        def __set_in_arcs(self, in_arcs):
            self.__in_arcs = in_arcs

        def __repr__(self):
            return str(self.name)

        def __eq__(self, other):
            # keep the ID for now in places
            return id(self) == id(other)

        def __hash__(self):
            # keep the ID for now in places
            return id(self)

        def __deepcopy__(self, memodict={}):
            if id(self) in memodict:
                return memodict[id(self)]
            new_place = ObjectCentricPetriNet.Place(
                self.name, properties=self.properties)
            memodict[id(self)] = new_place
            for arc in self.in_arcs:
                new_arc = deepcopy(arc, memo=memodict)
                new_place.in_arcs.add(new_arc)
            for arc in self.out_arcs:
                new_arc = deepcopy(arc, memo=memodict)
                new_place.out_arcs.add(new_arc)
            return new_place

        object_type = property(__get_object_type)
        initial = property(__get_initial)
        final = property(__get_final)
        out_arcs = property(__get_out_arcs, __set_out_arcs)
        in_arcs = property(__get_in_arcs, __set_in_arcs)
        name = property(__get_name, __set_name)

    class Transition(object):
        def __init__(self, name, label=None, in_arcs=None, out_arcs=None, properties=None, silent=None):
            self.__name = name
            self.__label = None if label is None else label
            self.__in_arcs = set() if in_arcs is None else in_arcs
            self.__out_arcs = set() if out_arcs is None else out_arcs
            self.__silent = silent
            self.__properties = dict() if properties is None else properties

        def __set_name(self, name):
            self.__name = name

        def __get_name(self):
            return self.__name

        def __set_label(self, label):
            self.__label = label

        def __get_label(self):
            return self.__label

        def __get_out_arcs(self):
            return self.__out_arcs

        def __get_in_arcs(self):
            return self.__in_arcs

        def __get_properties(self):
            return self.__properties

        def __get_silent(self):
            return self.__silent

        def __repr__(self):
            if self.label is None:
                return str(self.name)
            else:
                return str(self.label)

        def __eq__(self, other):
            # keep the ID for now in transitions
            return id(self) == id(other)

        def __hash__(self):
            # keep the ID for now in transitions
            return id(self)

        def __deepcopy__(self, memodict={}):
            if id(self) in memodict:
                return memodict[id(self)]
            new_trans = ObjectCentricPetriNet.Transition(
                self.name, self.label, properties=self.properties)
            memodict[id(self)] = new_trans
            for arc in self.in_arcs:
                new_arc = deepcopy(arc, memo=memodict)
                new_trans.in_arcs.add(new_arc)
            for arc in self.out_arcs:
                new_arc = deepcopy(arc, memo=memodict)
                new_trans.out_arcs.add(new_arc)
            return new_trans

        name = property(__get_name, __set_name)
        label = property(__get_label, __set_label)
        in_arcs = property(__get_in_arcs)
        out_arcs = property(__get_out_arcs)
        properties = property(__get_properties)
        silent = property(__get_silent)

    class Arc(object):
        def __init__(self, source, target, variable=False, weight=1, properties=None):
            if type(source) is type(target):
                raise Exception('Petri nets are bipartite graphs!')
            self.__source = source
            self.__target = target
            self.__weight = weight
            self.__variable = variable
            self.__properties = dict() if properties is None else properties

        def __get_source(self):
            return self.__source

        def __set_source(self, source):
            self.__source = source

        def __set_weight(self, weight):
            self.__weight = weight

        def __get_weight(self):
            return self.__weight

        def __get_target(self):
            return self.__target

        def __set_target(self, target):
            self.__target = target

        def __get_variable(self):
            return self.__variable

        def __get_properties(self):
            return self.__properties

        def __repr__(self):
            if type(self.source) is ObjectCentricPetriNet.Transition:
                if self.source.label:
                    return "(t)" + str(self.source.label) + "->" + "(p)" + str(self.target.name)
                else:
                    return "(t)" + str(self.source.name) + "->" + "(p)" + str(self.target.name)
            if type(self.target) is ObjectCentricPetriNet.Transition:
                if self.target.label:
                    return "(p)" + str(self.source.name) + "->" + "(t)" + str(self.target.label)
                else:
                    return "(p)" + str(self.source.name) + "->" + "(t)" + str(self.target.name)

        def __hash__(self):
            return id(self)

        def __eq__(self, other):
            return self.source == other.source and self.target == other.target

        def __deepcopy__(self, memodict={}):
            if id(self) in memodict:
                return memodict[id(self)]
            new_source = memodict[id(self.source)] if id(self.source) in memodict else deepcopy(self.source,
                                                                                                memo=memodict)
            new_target = memodict[id(self.target)] if id(self.target) in memodict else deepcopy(self.target,
                                                                                                memo=memodict)
            memodict[id(self.source)] = new_source
            memodict[id(self.target)] = new_target
            new_arc = ObjectCentricPetriNet.Arc(
                new_source, new_target, weight=self.weight, properties=self.properties)
            memodict[id(self)] = new_arc
            return new_arc

        source = property(__get_source, __set_source)
        target = property(__get_target, __set_target)
        variable = property(__get_variable)
        weight = property(__get_weight, __set_weight)
        properties = property(__get_properties)

    def __init__(self, name=None, places=None, transitions=None, arcs=None, properties=None, nets=None):
        self.__name = "" if name is None else name
        self.__places = places if places != None else set()
        self.__transitions = transitions if transitions != None else set()
        self.__arcs = arcs if arcs != None else set()
        self.__properties = dict() if properties is None else properties
        self.__nets = nets if nets is not None else dict()

    @property
    def name(self):
        return self.__name

    @name.setter
    def name(self, name):
        self.__name = name

    @property
    def places(self):
        return self.__places

    @property
    def transitions(self):
        return self.__transitions

    @property
    def arcs(self):
        return self.__arcs

    @property
    def properties(self):
        return self.__properties

    @property
    def object_types(self):
        return list(set([pl.object_type for pl in self.__places]))

    @property
    def nets(self):
        return self.__nets


@dataclass
class Marking(object):
    _tokens: Set[Tuple[ObjectCentricPetriNet.Place, str]
                 ] = field(default_factory=set)

    @property
    def tokens(self) -> Set[Tuple[ObjectCentricPetriNet.Place, str]]:
        return self._tokens

    def add_token(self, pl, obj):
        temp_tokens = set([(pl, oi) for (pl, oi) in self._tokens if oi == obj])
        self._tokens -= temp_tokens
        self._tokens.add((pl, obj))



import json

import pandas as pd
from dateutil import parser
from lxml import etree, objectify
import dateutil
from jsonschema import validate
import jsonschema
from datetime import datetime


def validate_with_schema(file_path, schema_path):
    if "json" in file_path.lower():
        file_content = json.load(open(file_path, "rb"))
        schema_content = json.load(open(schema_path, "rb"))
        try:
            validate(instance=file_content, schema=schema_content)
            return True
        except jsonschema.exceptions.ValidationError as err:
            return False
        return False
    elif "xml" in file_path:
        try:
            import lxml
            xml_file = lxml.etree.parse(file_path)
            xml_validator = lxml.etree.XMLSchema(file=schema_path)
            is_valid = xml_validator.validate(xml_file)
            return is_valid
        except:
            return False


def apply(file_path, return_obj_df=None, parameters=None):
    if "xml" in file_path:
        return apply_xml(file_path, return_obj_df=return_obj_df, parameters=parameters)
    elif "json" in file_path.lower():
        return apply_json(file_path, return_obj_df=return_obj_df, parameters=parameters)


def parse_xml(value, tag_str_lower):
    if "float" in tag_str_lower:
        return float(value)
    elif "date" in tag_str_lower:
        return datetime.fromisoformat(value)
    return str(value)


def apply_xml(file_path, return_obj_df=None, parameters=None):
    if parameters is None:
        parameters = {}

    parser = etree.XMLParser(remove_comments=True)
    tree = objectify.parse(file_path, parser=parser)
    root = tree.getroot()

    eve_stream = []
    obj_stream = []
    obj_types = {}

    for child in root:
        if child.tag.lower().endswith("events"):
            for event in child:
                eve = {}
                for child2 in event:
                    if child2.get("key") == "id":
                        eve["event_id"] = child2.get("value")
                    elif child2.get("key") == "timestamp":
                        eve["event_timestamp"] = datetime.fromisoformat(
                            child2.get("value"))
                    elif child2.get("key") == "activity":
                        eve["event_activity"] = child2.get("value")
                    elif child2.get("key") == "omap":
                        omap = []
                        for child3 in child2:
                            omap.append(child3.get("value"))
                        eve["@@omap"] = omap
                    elif child2.get("key") == "vmap":
                        for child3 in child2:
                            eve["event_" + child3.get("key")] = parse_xml(
                                child3.get("value"), child3.tag.lower())
                eve_stream.append(eve)

        elif child.tag.lower().endswith("objects"):
            for object in child:
                obj = {}
                for child2 in object:
                    if child2.get("key") == "id":
                        obj["object_id"] = child2.get("value")
                    elif child2.get("key") == "type":
                        obj["object_type"] = child2.get("value")
                    elif child2.get("key") == "ovmap":
                        for child3 in child2:
                            obj["object_" + child3.get("key")] = parse_xml(
                                child3.get("value"), child3.tag.lower())
                obj_stream.append(obj)

    for obj in obj_stream:
        obj_types[obj["object_id"]] = obj["object_type"]

    for eve in eve_stream:
        for obj in eve["@@omap"]:
            ot = obj_types[obj]
            if not ot in eve:
                eve[ot] = []
            eve[ot].append(obj)
        del eve["@@omap"]

    eve_df = pd.DataFrame(eve_stream)
    obj_df = pd.DataFrame(obj_stream)

    eve_df.type = "succint"

    if return_obj_df or (return_obj_df is None and len(obj_df.columns) > 1):
        return eve_df, obj_df
    return eve_df


def apply_json(file_path, return_obj_df=None, parameters=None):
    if parameters is None:
        parameters = {}

    prefix = "ocel:"
    F = open(file_path, "rb")
    obj = json.load(F)
    F.close()
    eve_stream = obj[prefix + "events"]
    for el in eve_stream:
        eve_stream[el]["event_id"] = el
    obj_stream = obj[prefix + "objects"]
    for el in obj_stream:
        obj_stream[el]["object_id"] = el
    obj_stream = list(obj_stream.values())
    obj_type = {}
    for el in obj_stream:
        obj_type[el["object_id"]] = el[prefix + "type"]
        el["object_type"] = el[prefix + "type"]
        del el[prefix + "type"]
        for k2 in el[prefix + "ovmap"]:
            el["object_" + k2] = el[prefix + "ovmap"][k2]
        del el[prefix + "ovmap"]
    eve_stream = list(eve_stream.values())
    for el in eve_stream:
        new_omap = {}
        for obj in el[prefix + "omap"]:
            typ = obj_type[obj]
            if not typ in new_omap:
                new_omap[typ] = set()
            new_omap[typ].add(obj)
        for typ in new_omap:
            new_omap[typ] = list(new_omap[typ])
        el[prefix + "omap"] = new_omap
        el["event_activity"] = el[prefix + "activity"]
        el["event_timestamp"] = datetime.fromisoformat(
            el[prefix + "timestamp"])
        del el[prefix + "activity"]
        del el[prefix + "timestamp"]
        for k2 in el[prefix + "vmap"]:
            el["event_" + k2] = el[prefix + "vmap"][k2]
        del el[prefix + "vmap"]
        for k2 in el[prefix + "omap"]:
            el[k2] = el[prefix + "omap"][k2]
        del el[prefix + "omap"]

    eve_df = pd.DataFrame(eve_stream)
    obj_df = pd.DataFrame(obj_stream)

    eve_df.type = "succint"

    if return_obj_df or (return_obj_df is None and len(obj_df.columns) > 1):
        return eve_df, obj_df
    return eve_df



from typing import Dict, List


class VmapParameters:
    def __init__(self, vmap_params, vmap_availables=None):
        self.vmap_params = vmap_params


class JsonParseParameters(VmapParameters):
    # Mapping from internal event keys to data internal keys
    event_params: Dict[str, str]
    # Mapping from internal object keys to data internal keys
    obj_params: Dict[str, str]
    # Mapping from internal log keys to data internal keys
    log_params: Dict[str, str]

    def __init__(self, vmap_params, vmap_availables):
        super().__init__(vmap_params, vmap_availables)
        self.event_params = {'act': 'ocel:activity',
                             'time': 'ocel:timestamp',
                             'omap': 'ocel:omap',
                             'vmap': 'ocel:vmap'}
        self.obj_params = {'type': 'ocel:type',
                           'ovmap': 'ocel:ovmap'}
        self.log_params = {'attr_names': 'ocel:attribute-names',
                           'obj_types': 'ocel:object-types',
                           'ordering': 'ocel:ordering',
                           'events': 'ocel:events',
                           'objects': 'ocel:objects',
                           'meta': 'ocel:global-log'}


class CsvParseParameters(VmapParameters):
    obj_names: List[str]
    val_names: List[str]
    time_name: str
    act_name: str

    def __init__(self, obj_names, val_names, time_name, act_name, vmap_params, vmap_availables):
        super().__init__(vmap_params, vmap_availables)
        self.obj_names = obj_names
        self.val_names = val_names
        self.time_name = time_name
        self.act_name = act_name



import ocpa.algo
import ocpa.objects
import ocpa.util
import ocpa.visualization

__version__ = '0.0.10'
__doc__ = "Object-Centric Process Analysis"
__author__ = 'Gyunam Park and Niklas Adams'
__author_email__ = 'gnpark@pads.rwth-aachen.de'
__maintainer__ = 'Gyunam Park and Niklas Adams'
__maintainer_email__ = "gnpark@pads.rwth-aachen.de"



import uuid
import tempfile
from graphviz import Digraph
from pm4py.objects.petri.petrinet import PetriNet
from ocpa.objects.oc_petri_net.obj import ObjectCentricPetriNet
from ocpa.visualization.oc_petri_net.util.util import equal_arc
from statistics import median, mean

COLORS = ["#05B202", "#A13CCD", "#BA0D39", "#39F6C0", "#E90638", "#07B423", "#306A8A", "#678225", "#2742FE", "#4C9A75",
          "#4C36E9", "#7DB022", "#EDAC54", "#EAC439", "#EAC439", "#1A9C45", "#8A51C4", "#496A63", "#FB9543", "#2B49DD",
          "#13ADA5", "#2DD8C1", "#2E53D7", "#EF9B77", "#06924F", "#AC2C4D", "#82193F", "#0140D3"]


def apply(ocpn, diagnostics, parameters=None):
    if parameters is None:
        parameters = {}

    image_format = "png"
    if "format" in parameters:
        image_format = parameters["format"]

    filename = tempfile.NamedTemporaryFile(suffix='.gv').name
    g = Digraph("", filename=filename, engine='dot',
                graph_attr={'bgcolor': 'transparent'})

    all_objs = {}
    trans_names = {}

    act_count = diagnostics["replayed_act_freq"]
    replayed_performance_min = diagnostics["replayed_performance_min"]
    replayed_performance_median = diagnostics[
        "replayed_performance_median"]
    replayed_performance_mean = diagnostics["replayed_performance_mean"]
    replayed_performance_max = diagnostics["replayed_performance_max"]

    agg_merged_group_size_hist = diagnostics["agg_merged_group_size_hist"]

    replayed_place_fitness = diagnostics["replayed_place_fitness"]

    replayed_arc_frequency = diagnostics["replayed_arc_frequency"]

    if 'act_count' in parameters:
        d_act_freq = parameters['act_count']
    else:
        d_act_freq = False

    if 'max_group_size' in parameters:
        d_max_group_size = parameters['max_group_size']
    else:
        d_max_group_size = False

    if 'min_group_size' in parameters:
        d_min_group_size = parameters['min_group_size']
    else:
        d_min_group_size = False

    if 'mean_group_size' in parameters:
        d_mean_group_size = parameters['mean_group_size']
    else:
        d_mean_group_size = False

    if 'median_group_size' in parameters:
        d_median_group_size = parameters['median_group_size']
    else:
        d_median_group_size = False

    if 'produced_token' in parameters:
        d_produced_token = parameters['produced_token']
    else:
        d_produced_token = False

    if 'consumed_token' in parameters:
        d_consumed_token = parameters['consumed_token']
    else:
        d_consumed_token = False

    if 'missing_token' in parameters:
        d_missing_token = parameters['missing_token']
    else:
        d_missing_token = False

    if 'remaining_token' in parameters:
        d_remaining_token = parameters['remaining_token']
    else:
        d_remaining_token = False

    if 'arc_freq' in parameters:
        d_arc_freq = parameters['arc_freq']
    else:
        d_arc_freq = False

    if 'avg_sojourn_time' in parameters:
        d_avg_sojourn_time = parameters['avg_sojourn_time']
    else:
        d_avg_sojourn_time = False

    if 'med_sojourn_time' in parameters:
        d_med_sojourn_time = parameters['med_sojourn_time']
    else:
        d_med_sojourn_time = False

    if 'min_sojourn_time' in parameters:
        d_min_sojourn_time = parameters['min_sojourn_time']
    else:
        d_min_sojourn_time = False

    if 'max_sojourn_time' in parameters:
        d_max_sojourn_time = parameters['max_sojourn_time']
    else:
        d_max_sojourn_time = False

    pl_count = 1
    tr_count = 1
    arc_count = 1

    # color = COLORS[index % len(COLORS)]
    color = "#05B202"
    color_mapping = dict()
    object_types = ocpn.object_types
    for index, ot in enumerate(object_types):
        color_mapping[ot] = COLORS[index % len(COLORS)]

    for pl in ocpn.places:
        # this_uuid = str(uuid.uuid4())
        this_uuid = "p%d" % (pl_count)
        # pl_str = this_uuid
        pl_count += 1
        color = color_mapping[pl.object_type]

        pl_label = pl.name
        if d_missing_token == True:
            pl_str = "p=%d r=%d\\nc=%d m=%d" % (
                replayed_place_fitness[pl.name]["p"], replayed_place_fitness[pl.name]["r"], replayed_place_fitness[pl.name]["c"], replayed_place_fitness[pl.name]["m"])
            pl_label += "\n %s" % (pl_str)

        if pl.initial == True:
            g.node("(p)%s" % (pl.name), pl.name, shape="circle", style="filled", fillcolor=color, color=color,
                   fontsize="13.0", labelfontsize="13.0")
        elif pl.final == True:
            g.node("(p)%s" % (pl.name), pl.name, shape="circle", style="filled", color=color, fillcolor=color,
                   fontsize="13.0", labelfontsize="13.0")
        else:
            g.node("(p)%s" % (pl.name), pl_label, shape="circle", color=color,
                   fontsize="13.0", labelfontsize="13.0")
        all_objs[pl] = "(p)%s" % (pl.name)

    for tr in ocpn.transitions:
        # this_uuid = str(uuid.uuid4())
        this_uuid = "t%d" % (tr_count)
        tr_count += 1
        if tr.silent == True:
            g.node(this_uuid, this_uuid, fontcolor="#FFFFFF", shape="box",
                   fillcolor="#000000", style="filled", labelfontsize="18.0")
            all_objs[tr] = this_uuid
        elif tr.name not in trans_names:
            tr_label = tr.name
            tr_xlabel = "-"
            if d_act_freq == True:
                tr_label += "\n (%d)" % (act_count[tr.name])

            if d_max_group_size == True:
                tr_xlabel += "\n max: "
                for obj_type in agg_merged_group_size_hist[tr.name]["max"].keys():
                    tr_xlabel += "%s=%d " % (obj_type,
                                             agg_merged_group_size_hist[tr.name]["max"][obj_type])

            if d_min_group_size == True:
                tr_xlabel += "\n min: "
                for obj_type in agg_merged_group_size_hist[tr.name]["min"].keys():
                    tr_xlabel += "%s=%d " % (obj_type,
                                             agg_merged_group_size_hist[tr.name]["min"][obj_type])

            if d_mean_group_size == True:
                tr_xlabel += "\n mean: "
                for obj_type in agg_merged_group_size_hist[tr.name]["mean"].keys():
                    tr_xlabel += "%s=%d " % (
                        obj_type, agg_merged_group_size_hist[tr.name]["mean"][obj_type])

            if d_median_group_size == True:
                tr_xlabel += "\n median: "
                for obj_type in agg_merged_group_size_hist[tr.name]["median"].keys():
                    tr_xlabel += "%s=%d " % (
                        obj_type, agg_merged_group_size_hist[tr.name]["median"][obj_type])
            g.node("(t)%s" % (tr.name), "%s" % (tr_label), shape="box", fontsize="36.0",
                   labelfontsize="36.0", xlabel='''<<FONT POINT-SIZE="11">%s</FONT>>''' % (tr_xlabel))
            trans_names[tr.name] = tr.name
            all_objs[tr] = "(t)%s" % (tr.name)
        else:
            all_objs[tr] = trans_names[tr.name]

    for arc in ocpn.arcs:
        # this_uuid = str(uuid.uuid4())
        this_uuid = "a%d" % (arc_count)
        arc_count += 1
        arc_annotation = ""

        source_node = arc.source
        target_node = arc.target

        if type(source_node) == ObjectCentricPetriNet.Place:
            object_type = source_node.object_type
        elif type(target_node) == ObjectCentricPetriNet.Place:
            object_type = target_node.object_type

        color = color_mapping[object_type]

        if d_max_sojourn_time == True:
            if repr(arc) in replayed_performance_max.keys():
                arc_annotation += "\n max. sojourn time=%s" % (
                    replayed_performance_max[repr(arc)])

        if d_avg_sojourn_time == True:
            if repr(arc) in replayed_performance_mean.keys():
                arc_annotation += "\n avg. sojourn time=%s" % (
                    replayed_performance_mean[repr(arc)])

        if d_med_sojourn_time == True:
            if repr(arc) in replayed_performance_median.keys():
                arc_annotation += "\n med. sojourn time=%s" % (
                    replayed_performance_median[repr(arc)])

        if d_min_sojourn_time == True:
            if repr(arc) in replayed_performance_min.keys():
                arc_annotation += "\n min. sojourn time=%s" % (
                    replayed_performance_min[repr(arc)])

        penwidth = "1"
        if d_arc_freq == True:
            arc_freq = replayed_arc_frequency[arc.__repr__()]
            arc_annotation += "\n freq=%s" % (arc_freq)
        if arc.variable == True:
            g.edge(all_objs[source_node], all_objs[target_node],
                   label=arc_annotation, color=color + ":white:" + color, fontsize="13.0", penwidth=penwidth)
        else:
            g.edge(all_objs[source_node], all_objs[target_node],
                   label=arc_annotation, color=color, fontsize="13.0", penwidth=penwidth)

        all_objs[arc] = this_uuid

    g.attr(overlap='false')
    g.attr(fontsize='11')

    g.format = image_format
    return g



import pandas as pd
from copy import deepcopy
from pm4py.util.constants import PARAMETER_CONSTANT_CASEID_KEY, PARAMETER_CONSTANT_ATTRIBUTE_KEY
from pm4py.algo.discovery.dfg.adapters.pandas import df_statistics


def apply(all_df, return_obj_dataframe=False, parameters=None):
    if parameters is None:
        parameters = {}

    eve_cols = [x for x in all_df.columns if not x.startswith("object_")]
    obj_cols = [x for x in all_df.columns if x.startswith("object_")]
    df = all_df[eve_cols]
    obj_df = pd.DataFrame()
    if obj_cols:
        obj_df = all_df[obj_cols]
    df["event_timestamp"] = pd.to_datetime(df["event_timestamp"])
    if "event_start_timestamp" in df.columns:
        df["event_start_timestamp"] = pd.to_datetime(
            df["event_start_timestamp"])
    df = df.dropna(subset=["event_id"])
    df["event_id"] = df["event_id"].astype(str)
    df.type = "succint"

    if return_obj_dataframe:
        obj_df = obj_df.dropna(subset=["object_id"])
        return df, obj_df

    return df


def filter_by_timestamp(df, start_timestamp=None, end_timestamp=None):
    if start_timestamp is not None:
        df = df.loc[df["event_timestamp"] >= start_timestamp]
    if end_timestamp is not None:
        df = df.loc[df["event_timestamp"] <= end_timestamp]
    return df


def filter_object_df_by_object_ids(df, ids):
    df = df.loc[df["object_id"].isin(ids)]
    return df



def succint_stream_to_exploded_stream(stream):
    new_stream = []

    for ev in stream:
        keys = set(ev.keys())

        event_keys = [k for k in keys if k.startswith("event_")]
        object_keys = [k for k in keys if not k in event_keys]

        basic_event = {k: ev[k] for k in event_keys}

        for k in object_keys:
            if type(ev[k]) is str:
                if ev[k][0] == "[":
                    ev[k] = eval(ev[k])
                    #ev[k] = ev[k][1:-1].split(",")
            values = ev[k]
            if values is not None:
            #if values is not None and len(values) > 0:
                if not (str(values).lower() == "nan" or str(values).lower() == "nat"):
                    for v in values:
                        event = deepcopy(basic_event)
                        event[k] = v
                        new_stream.append(event)

    return new_stream

def succint_mdl_to_exploded_mdl(df):
    stream = df.to_dict('r')

    exploded_stream = succint_stream_to_exploded_stream(stream)

    df = pd.DataFrame(exploded_stream)
    df.type = "exploded"

    return df

def clean_frequency(df, min_acti_freq=0):
    try:
        if df.type == "succint":
            df = succint_mdl_to_exploded_mdl(df)
    except:
        pass
    activ = dict(df.groupby("event_id").first()["event_activity"].value_counts())
    activ = [x for x,y in activ.items() if y >= min_acti_freq]
    return df[df["event_activity"].isin(activ)]

def filter_paths(df, paths, parameters=None):
    """
    Apply a filter on traces containing / not containing a path

    Parameters
    ----------
    df
        Dataframe
    paths
        Paths to filter on
    parameters
        Possible parameters of the algorithm, including:
            case_id_glue -> Case ID column in the dataframe
            attribute_key -> Attribute we want to filter
            positive -> Specifies if the filter should be applied including traces (positive=True)
            or excluding traces (positive=False)
    Returns
    ----------
    df
        Filtered dataframe
    """
    try:
        if df.type == "succint":
            df = succint_mdl_to_exploded_mdl(df)
    except:
        pass
    if parameters is None:
        parameters = {}
    paths = [path[0] + "," + path[1] for path in paths]
    case_id_glue = parameters[
        PARAMETER_CONSTANT_CASEID_KEY] if PARAMETER_CONSTANT_CASEID_KEY in parameters else CASE_CONCEPT_NAME
    attribute_key = parameters[
        PARAMETER_CONSTANT_ATTRIBUTE_KEY] if PARAMETER_CONSTANT_ATTRIBUTE_KEY in parameters else DEFAULT_NAME_KEY
    df = df.sort_values([case_id_glue, "event_timestamp"])
    positive = parameters["positive"] if "positive" in parameters else True
    filt_df = df[[case_id_glue, attribute_key, "event_id"]]
    filt_dif_shifted = filt_df.shift(-1)
    filt_dif_shifted.columns = [str(col) + '_2' for col in filt_dif_shifted.columns]
    stacked_df = pd.concat([filt_df, filt_dif_shifted], axis=1)
    stacked_df["@@path"] = stacked_df[attribute_key] + "," + stacked_df[attribute_key + "_2"]
    stacked_df = stacked_df[stacked_df["@@path"].isin(paths)]
    i1 = df.set_index("event_id").index
    i2 = stacked_df.set_index("event_id").index
    i3 = stacked_df.set_index("event_id_2").index
    if positive:
        return df[i1.isin(i2) | i1.isin(i3)]
    else:
        return df[~i1.isin(i2) & ~i1.isin(i3)]


def clean_arc_frequency(df, min_freq=0):
    if min_freq > 0:
        persps = [x for x in df.columns if not x.startswith("event_")]
        collation = []
        for persp in persps:
            red_df = df.dropna(subset=[persp])
            prevlen = len(df)
            while True:
                dfg = df_statistics.get_dfg_graph(red_df, activity_key="event_activity", timestamp_key="event_timestamp", case_id_glue=persp)
                dfg = [x for x in dfg if dfg[x] >= min_freq]
                param = {}
                param[PARAMETER_CONSTANT_CASEID_KEY] = persp
                param[PARAMETER_CONSTANT_ATTRIBUTE_KEY] = "event_activity"
                red_df = filter_paths(red_df, dfg, parameters=param)
                thislen = len(red_df)
                dfg = df_statistics.get_dfg_graph(red_df, activity_key="event_activity", timestamp_key="event_timestamp", case_id_glue=persp)
                if len(dfg) == 0 or min(dfg.values()) >= min_freq or prevlen == thislen:
                    collation.append(red_df)
                    break
                prevlen = thislen
        return pd.concat(collation)
    return df



from ocpa.objects.log.converter.versions import jsonocel_to_mdl

JSONOCEL_TO_MDL = "json_to_mdl"

VERSIONS = {JSONOCEL_TO_MDL: jsonocel_to_mdl.apply}


def apply(ocel, variant=JSONOCEL_TO_MDL, parameters=None):
    return VERSIONS[variant](ocel, parameters=parameters)



import json

import pandas as pd
from dateutil import parser
from lxml import etree, objectify
import dateutil
from jsonschema import validate
import jsonschema
from datetime import datetime


def apply(ocel, return_obj_df=True, parameters=None):
    if parameters is None:
        parameters = {}

    prefix = "ocel:"

    objects = ocel.raw.objects
    events = ocel.raw.events

    obj_type = {}
    for obj in objects:
        obj_type[objects[obj].id] = objects[obj].type

    eve_stream = []
    for ev in events:
        new_omap = {}
        for obj in events[ev].omap:
            typ = obj_type[obj]
            if not typ in new_omap:
                new_omap[typ] = set()
            new_omap[typ].add(obj)
        for typ in new_omap:
            new_omap[typ] = list(new_omap[typ])
        el = {}
        el["event_id"] = events[ev].id
        el["event_activity"] = events[ev].act
        el["event_timestamp"] = events[ev].time
        for k2 in events[ev].vmap:
            el["event_" + k2] = events[ev].vmap[k2]
        for k2 in new_omap:
            el[k2] = new_omap[k2]
        eve_stream.append(el)

    obj_stream = []

    eve_df = pd.DataFrame(eve_stream)
    obj_df = pd.DataFrame(obj_stream)

    eve_df.type = "succint"

    if return_obj_df or (return_obj_df is None and len(obj_df.columns) > 1):
        return eve_df, obj_df
    return eve_df



from setuptools import setup

import ocpa
setup(
    name = ocpa.__name__,
    packages = [
        'ocpa', 'ocpa.algo', 'ocpa.objects', 'ocpa.util', 'ocpa.visualization',
        'ocpa.algo.discovery', 'ocpa.algo.discovery.mvp', 'ocpa.algo.discovery.ocpn', 'ocpa.algo.discovery.mvp.projection', 'ocpa.algo.discovery.mvp.projection.versions', 'ocpa.algo.discovery.ocpn.versions', 
        'ocpa.algo.evaluation', 'ocpa.algo.evaluation.precision',
        'ocpa.objects.log', 'ocpa.objects.log.converter', 'ocpa.objects.log.converter.versions', 'ocpa.objects.log.importer.mdl', 'ocpa.objects.log.importer.ocel', 'ocpa.objects.log.importer.ocel.versions', 'ocpa.objects.log.util', 'ocpa.objects.oc_petri_net', 
        'ocpa.util', 
        'ocpa.visualization', 'ocpa.visualization.oc_petri_net', 'ocpa.visualization.oc_petri_net.util', 'ocpa.visualization.oc_petri_net.versions'
        ],
    py_modules=[ocpa.__name__],
    include_package_data=True,
    version = ocpa.__version__,      
    license='MIT',        # Chose a license from here: https://help.github.com/articles/licensing-a-repository
    description = ocpa.__doc__.strip(),   
    author=ocpa.__author__,
    author_email=ocpa.__author_email__,
    url = 'https://github.com/gyunamister/ocpa',   # Provide either the link to your github or to your website
    install_requires=[            # I get to this in a second
            "pm4py>=2.2.0",
            "setuptools",
            "jsonschema",
        ],
    # classifiers=[
    # 'Development Status :: 3 - Alpha',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
    # 'Intended Audience :: Developers',      # Define that your audience are developers
    # 'Topic :: Software Development :: Build Tools',
    # 'License :: OSI Approved :: MIT License',   # Again, pick a license
    # 'Programming Language :: Python :: 3',      #Specify which pyhton versions that you want to support
    # 'Programming Language :: Python :: 3.4',
    # 'Programming Language :: Python :: 3.5',
    # 'Programming Language :: Python :: 3.6',
    # ],
)



from dataclasses import dataclass, field
from typing import List, Dict, Set, Any, Optional, Union, Tuple
from datetime import datetime


from ocpa.objects.log.util.param import CsvParseParameters, JsonParseParameters


@dataclass
class EventId:
    id: str


@dataclass
class EventClassic(EventId):
    act: str
    time: datetime


@dataclass
class EventClassicResource(EventClassic):
    vmap: Dict[str, Any]


@dataclass
class Event(EventClassic):
    omap: List[str]
    vmap: Dict[str, Any]
    # Kept for backward compatibility with the evaluation
    corr: bool = field(default_factory=lambda: False)


@dataclass
class Obj:
    id: str
    type: str
    ovmap: Dict


@dataclass
class MetaObjectCentricData:
    attr_names: List[str]  # AN
    attr_types: List[str]  # AT
    attr_typ: Dict  # pi_typ

    obj_types: List[str]  # OT

    act_attr: Dict[str, List[str]]  # allowed attr per act
    # act_obj: Dict[str, List[str]]  # allowed ot per act

    acts: Set[str] = field(init=False)  # TODO: change to list for json
    ress: Set[str] = field(init=False)  # TODO: change to list for json
    locs: Set[str] = field(init=False)  # TODO: change to list for json
    # Used for OCEL json data to simplify UI on homepage
    attr_events: List[str] = field(default_factory=lambda: [])

    def __post_init__(self):
        self.acts = {act for act in self.act_attr}


@dataclass
class RawObjectCentricData:
    events: Dict[str, Event]
    objects: Dict[str, Obj]

    @property
    def obj_ids(self) -> List[str]:
        return list(self.objects.keys())


@dataclass
class ObjectCentricData:
    meta: MetaObjectCentricData
    raw: RawObjectCentricData
    vmap_param: Union[CsvParseParameters, JsonParseParameters]

    def __post_init__(self):
        self.meta.locs = {}


def sort_events(data: ObjectCentricData) -> None:
    events = data.raw.events
    data.raw.events = {k: event for k, event in sorted(
        events.items(), key=lambda item: item[1].time)}
